# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - ItemSupport library

    @copyright: 2013, 2014 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin import config
from MoinMoin.Page import Page
from MoinMoin.PageEditor import PageEditor
from MoinMoin.security import Permissions
from MoinMoin.util import lock
from MoinSupport import getMetadata, getPagesForSearch
import os

# Content storage support.

class SpecialPermissionsForPage(Permissions):

    "Permit saving of ACL-enabled pages."

    def __init__(self, user, pagename):
        Permissions.__init__(self, user)
        self.pagename = pagename

    def admin(self, pagename):
        return pagename == self.pagename

    write = admin

class ReadPermissionsForSubpages(Permissions):

    "Permit listing of ACL-affected subpages."

    def __init__(self, user, pagename):
        Permissions.__init__(self, user)
        self.pagename = pagename

    def read(self, pagename):
        return pagename.startswith("%s/" % self.pagename)

# Underlying storage mechanisms.

class GeneralItemStore:

    "Common item store functionality."

    def __init__(self, lock_dir):

        "Initialise an item store with the given 'lock_dir' guarding access."

        self.lock_dir = lock_dir
        self.writelock = lock.WriteLock(lock_dir)
        self.readlock = lock.ReadLock(lock_dir)

    # High-level methods.

    def __len__(self):

        """
        Return the number of items.
        """

        return len(self.keys())

    def __iter__(self):

        "Return an iterator over the items in the store."

        return ItemIterator(self)

    def keys(self):

        "Return a list of keys for items in the store."

        self.readlock.acquire()
        try:
            return self.get_keys()
        finally:
            self.readlock.release()

    def __getitem__(self, number):

        "Return the item with the given 'number'."

        self.readlock.acquire()
        try:
            try:
                return self.read_item(number)
            except (IOError, OSError):
                raise IndexError, number
        finally:
            self.readlock.release()

    def __delitem__(self, number):

        "Remove the item with the given 'number' from the store."

        self.writelock.acquire()
        try:
            try:
                self.remove_item(number)
            except (IOError, OSError):
                raise IndexError, number
        finally:
            self.writelock.release()

    def next(self):

        """
        Return the number of the next item (which should also be the number of
        items if none have been deleted).
        """

        self.writelock.acquire()
        try:
            return self.get_next()
        finally:
            self.writelock.release()

class SequentialAccess(GeneralItemStore):

    "Support sequential access to items."

    def deduce_next(self):

        "Deduce the next item number from the existing item files."

        return max(self.get_keys() or [-1]) + 1

    def read_next(self):

        "Read the next item number from a special file."

        if not os.path.exists(self.next_path):
            return None

        f = open(self.next_path)
        try:
            try:
                return int(f.read())
            except ValueError:
                return None
        finally:
            f.close()

    def write_next(self, next):

        "Write the 'next' item number to a special file."

        f = open(self.next_path, "w")
        try:
            f.write(str(next))
        finally:
            f.close()

class DirectoryStore(GeneralItemStore):

    "A directory-based item store."

    def __init__(self, path, lock_dir):

        "Initialise an item store for the given 'path' and 'lock_dir'."

        self.path = path
        self.next_path = os.path.join(self.path, "next")
        self.lock_dir = lock_dir
        self.writelock = lock.WriteLock(lock_dir)
        self.readlock = lock.ReadLock(lock_dir)

    def mtime(self):

        "Return the last modified time of the item store directory."

        return os.path.getmtime(self.path)

    def get_keys(self):

        "Return the item keys."

        return [int(filename) for filename in os.listdir(self.path) if filename.isdigit()]

    def write_item(self, item, next):

        "Write the given 'item' to a file with the given 'next' item number."

        f = open(self.get_item_path(next), "wb")
        try:
            f.write(item)
        finally:
            f.close()

    def read_item(self, identifier):

        "Read the item with the given item 'identifier'."

        f = open(self.get_item_path(identifier), "rb")
        try:
            return f.read()
        finally:
            f.close()

    def remove_item(self, identifier):

        "Remove the item with the given item 'identifier'."

        os.remove(self.get_item_path(identifier))

    def get_item_path(self, identifier):

        "Get the path for the given item 'identifier'."

        if isinstance(identifier, unicode):
            filename = identifier.encode(config.charset)
        else:
            filename = identifier

        path = os.path.abspath(os.path.join(self.path, filename))
        basepath = os.path.join(self.path, "")

        if os.path.commonprefix([path, basepath]) != basepath:
            raise OSError, path

        return path

class DirectoryItemStore(DirectoryStore, SequentialAccess):

    "A directory-based item store with numeric keys."

    def get_next(self):

        "Return the next item number."

        next = self.read_next()
        if next is None:
            next = self.deduce_next()
            self.write_next(next)
        return next

    # High-level methods.

    def append(self, item):

        "Append the given 'item' to the store."

        self.writelock.acquire()
        try:
            next = self.get_next()
            self.write_item(item, next)
            self.write_next(next + 1)
        finally:
            self.writelock.release()

class DirectoryNamedItemStore(DirectoryStore):

    "A directory-based item store with explicit keys."

    def __setitem__(self, name, item):

        "Using the given 'name', set the given 'item' in the store."

        self.writelock.acquire()
        try:
            self.write_item(item, name)
        finally:
            self.writelock.release()

class SubpageItemStore(SequentialAccess):

    "A subpage-based item store."

    def __init__(self, page, lock_dir):

        "Initialise an item store for subpages under the given 'page'."

        GeneralItemStore.__init__(self, lock_dir)
        self.page = page

    def mtime(self):

        "Return the last modified time of the item store."

        keys = self.get_keys()
        if not keys:
            page = self.page
        else:
            page = Page(self.page.request, self.get_item_path(max(keys)))

        return wikiutil.version2timestamp(
            getMetadata(page)["last-modified"]
            )

    def get_next(self):

        "Return the next item number."

        return self.deduce_next()

    def get_keys(self):

        "Return the item keys."

        request = self.page.request

        # Collect the strict subpages of the parent page.

        leafnames = []
        parentname = self.page.page_name

        # To list pages whose ACLs may prevent access, a special policy is required.

        may = request.user.may
        request.user.may = ReadPermissionsForSubpages(request.user, parentname)

        try:
            for page in getPagesForSearch("title:regex:^%s/" % parentname, self.page.request):
                basename, leafname = page.page_name.rsplit("/", 1)

                # Only collect numbered pages immediately below the parent.

                if basename == parentname and leafname.isdigit():
                    leafnames.append(int(leafname))

            return leafnames

        # Restore the original policy.

        finally:
            request.user.may = may

    def write_item(self, item, next):

        "Write the given 'item' to a page with the given 'next' item number."

        request = self.page.request
        pagename = self.get_item_path(next)

        # To add a page with an ACL, a special policy is required.

        may = request.user.may
        request.user.may = SpecialPermissionsForPage(request.user, pagename)

        # Attempt to save the page, copying any ACL.

        try:
            page = PageEditor(request, pagename)
            page.saveText(item, 0)

        # Restore the original policy.

        finally:
            request.user.may = may

    def read_item(self, number):

        "Read the item with the given item 'number'."

        page = Page(self.page.request, self.get_item_path(number))
        return page.get_raw_body()

    def remove_item(self, number):

        "Remove the item with the given item 'number'."

        page = PageEditor(self.page.request, self.get_item_path(number))
        page.deletePage()

    def get_item_path(self, number):

        "Get the path for the given item 'number'."

        return "%s/%s" % (self.page.page_name, number)

    # High-level methods.

    def append(self, item):

        "Append the given 'item' to the store."

        self.writelock.acquire()
        try:
            next = self.get_next()
            self.write_item(item, next)
        finally:
            self.writelock.release()

class ItemIterator:

    "An iterator over items in a store."

    def __init__(self, store, direction=1, keys=None):
        self.store = store
        self.direction = direction
        self.keys = keys
        self.reset()

    def reset(self):
        if self.direction == 1:
            self._next = 0
            self.final = self.store.next()
        else:
            self._next = self.store.next() - 1
            self.final = 0

    def more(self):
        if self.direction == 1:
            return self._next < self.final
        else:
            return self._next >= self.final

    def get_next(self):
        if self.keys:
            next = self.keys[self._next]
        else:
            next = self._next
        self._next += self.direction
        return next

    def next(self):
        while self.more():
            try:
                return self.store[self.get_next()]
            except IndexError:
                pass

        raise StopIteration

    def reverse(self):
        self.direction = -self.direction
        self.reset()

    def reversed(self):
        self.reverse()
        return self

    def __iter__(self):
        return self

def getDirectoryItemStoreForPage(page, item_dir, lock_dir):

    """
    A convenience function returning a directory-based store for the given
    'page', using the given 'item_dir' and 'lock_dir'.
    """

    item_dir_path = tuple(item_dir.split("/"))
    lock_dir_path = tuple(lock_dir.split("/"))
    return DirectoryItemStore(page.getPagePath(*item_dir_path), page.getPagePath(*lock_dir_path))

def getDirectoryNamedItemStoreForPage(page, item_dir, lock_dir):

    """
    A convenience function returning a directory-based store for the given
    'page', using the given 'item_dir' and 'lock_dir'.
    """

    item_dir_path = tuple(item_dir.split("/"))
    lock_dir_path = tuple(lock_dir.split("/"))
    return DirectoryNamedItemStore(page.getPagePath(*item_dir_path), page.getPagePath(*lock_dir_path))

def getSubpageItemStoreForPage(page, lock_dir):

    """
    A convenience function returning a subpage-based store for the given
    'page', using the given 'lock_dir'.
    """

    lock_dir_path = tuple(lock_dir.split("/"))
    return SubpageItemStore(page, page.getPagePath(*lock_dir_path))

# Page-oriented item store classes.

class ItemStoreBase:

    "Access item stores via pages, observing page access restrictions."

    def __init__(self, page, store):
        self.page = page
        self.store = store

    def can_write(self):

        """
        Return whether the user associated with the request can write to the
        page owning this store.
        """

        user = self.page.request.user
        return user and user.may.write(self.page.page_name)

    def can_read(self):

        """
        Return whether the user associated with the request can read from the
        page owning this store.
        """

        user = self.page.request.user
        return user and user.may.read(self.page.page_name)

    def can_delete(self):

        """
        Return whether the user associated with the request can delete the
        page owning this store.
        """

        user = self.page.request.user
        return user and user.may.delete(self.page.page_name)

    # Store-specific methods.

    def mtime(self):
        return self.store.mtime()

    # High-level methods.

    def keys(self):

        "Return a list of keys for items in the store."

        if not self.can_read():
            return 0

        return self.store.keys()

    def __len__(self):

        "Return the number of items in the store."

        if not self.can_read():
            return 0

        return len(self.store)

    def __getitem__(self, number):

        "Return the item with the given 'number'."

        if not self.can_read():
            raise IndexError, number

        return self.store.__getitem__(number)

    def __delitem__(self, number):

        "Remove the item with the given 'number'."

        if not self.can_delete():
            return

        return self.store.__delitem__(number)

    def __iter__(self):
        return self.store.__iter__()

    def next(self):
        return self.store.next()

class SequentialStoreBase:

    "Sequential access methods for item stores."

    def append(self, item):

        "Append the given 'item' to the store."

        if not self.can_write():
            return

        self.store.append(item)

class NamedStoreBase:

    "Name-based access methods for item stores."

    def __setitem__(self, name, item):

        "Using the given 'name', set the given 'item' in the store."

        if not self.can_write():
            return

        self.store[name] = item

# Convenience store classes.

class ItemStore(ItemStoreBase, SequentialStoreBase):

    "Store items in a directory via a page."

    def __init__(self, page, item_dir="items", lock_dir=None):
        ItemStoreBase.__init__(self, page, getDirectoryItemStoreForPage(page, item_dir, lock_dir or ("%s-locks" % item_dir)))

class NamedItemStore(ItemStoreBase, NamedStoreBase):

    "Store items in a directory via a page."

    def __init__(self, page, item_dir="items", lock_dir=None):
        ItemStoreBase.__init__(self, page, getDirectoryNamedItemStoreForPage(page, item_dir, lock_dir or ("%s-locks" % item_dir)))

class ItemSubpageStore(ItemStoreBase):

    "Store items in subpages of a page."

    def __init__(self, page, lock_dir=None):
        ItemStoreBase.__init__(self, page, getSubpageItemStoreForPage(page, lock_dir or "subpage-items-locks"))

# vim: tabstop=4 expandtab shiftwidth=4
