# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - MoinRemoteSupport library

    @copyright: 2011, 2012, 2013, 2014 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from ContentTypeSupport import getContentTypeAndEncoding
from MoinMoin.action import cache
from MoinMoin import caching, log
from email.parser import Parser
from email.mime.multipart import MIMEMultipart
from urllib import splithost, splitpasswd, splitport, splituser, unquote_plus
from urlparse import urlsplit
import urllib2, time
import imaplib

logging = log.getLogger(__name__)

def getCachedResource(request, url, arena, scope, max_cache_age, reader=None):

    """
    Using the given 'request', return the resource data for the given 'url',
    accessing a cache entry with the given 'arena' and 'scope' where the data
    has already been downloaded. The 'max_cache_age' indicates the length in
    seconds that a cache entry remains valid.

    If the optional 'reader' object is given, it will be used to access the
    'url' and write the downloaded data to a cache entry. Otherwise, a standard
    URL reader will be used.

    If the resource cannot be downloaded and cached, None is returned.
    Otherwise, the form of the data is as follows:

    url <newline>
    [ content-type-header ] <newline>
    [ other-header <newline> ]*
    <newline>
    content-body
    """

    reader = reader or urlreader

    # See if the URL is cached.

    cache_key = cache.key(request, content=url)
    cache_entry = caching.CacheEntry(request, arena, cache_key, scope=scope)

    # If no entry exists, or if the entry is older than the specified age,
    # create one with the response from the URL.

    now = time.time()
    mtime = cache_entry.mtime()

    # NOTE: The URL could be checked and the 'If-Modified-Since' header
    # NOTE: (see MoinMoin.action.pollsistersites) could be checked.

    if not cache_entry.exists() or cache_entry.size() == 0 or now - mtime >= max_cache_age:

        # Access the remote data source.

        cache_entry.open(mode="w")

        try:
            try:
                # Read from the source and write to the cache.

                reader(url, cache_entry)

            # In case of an exception, return None.

            except IOError:
                if cache_entry.exists():
                    cache_entry.remove()
                return None

        finally:
            cache_entry.close()

    # Open the cache entry and read it.

    cache_entry.open()
    try:
        return cache_entry.read()
    finally:
        cache_entry.close()

def urlreader(url, cache_entry):

    "Retrieve data from the given 'url', writing it to the 'cache_entry'."

    f = urllib2.urlopen(url)
    try:
        writeCacheHeaders(url, f.headers, cache_entry)
        cache_entry.write(f.read())
    finally:
        f.close()

def imapreader(url, cache_entry):

    """
    Retrieve data associated with the given 'url' using the IMAP protocol
    specifically, writing it to the 'cache_entry'.
    """

    # NOTE: Should use something like pykolab.imap_utf7.encode here.

    enc = lambda s: s.encode("utf-7")

    # The URL maps to credentials and folder details.

    scheme, netloc, path, query, fragment = urlsplit(url)
    credentials, location = splituser(netloc)
    username, password = map(unquote_plus, splitpasswd(credentials))
    host, port = splitport(location)
    folders = map(unquote_plus, path.split("/")[1:])

    # Connect and log in to the IMAP server.

    cls = scheme == "imaps" and imaplib.IMAP4_SSL or imaplib.IMAP4

    if port is None:
        i = cls(host)
    else:
        i = cls(host, int(port))

    i.login(username, password)

    try:
        # Descend to the desired folder.

        for folder in folders:
            code, response = i.select(enc(folder), readonly=True)
            if code != "OK":
                logging.warning("Could not enter folder: %s" % folder)
                raise IOError

        # Search for all messages.
        # NOTE: This could also be parameterised.

        code, response = i.search(None, "(ALL)")

        if code != "OK":
            logging.warning("Could not enter folder: %s" % folder)
            raise IOError

        # For each result, obtain the full message, but embed it in a larger
        # multipart message.

        message = MIMEMultipart()

        writeCacheHeaders(url, message, cache_entry)

        numbers = response and response[0].split(" ") or []

        for n in numbers:
            code, response = i.fetch(n, "(RFC822.PEEK)")

            if code == "OK" and response:

                # Write the message payload into the cache entry for later
                # processing.

                for data in response:
                    try:
                        envelope, body = data
                        message.attach(Parser().parsestr(body))
                    except ValueError:
                        pass
            else:
                logging.warning("Could not obtain message %d from folder %s" % (n, folder))

        cache_entry.write(message.as_string())

    finally:
        i.logout()
        del i

def writeCacheHeaders(url, headers, cache_entry):

    """
    For the given 'url', write it and the given 'headers' to the given
    'cache_entry'.
    """

    cache_entry.write(url + "\n")
    cache_entry.write((headers.get("content-type") or "") + "\n")
    for key, value in headers.items():
        if key.lower() != "content-type":
            cache_entry.write("%s: %s\n" % (key, value))
    cache_entry.write("\n")

def getCachedResourceMetadata(f):

    "Return a metadata dictionary for the given resource file-like object 'f'."

    url = f.readline()
    content_type, encoding = getContentTypeAndEncoding(f.readline())

    metadata = {}
    line = f.readline()

    while line.strip():
        key, value = [v.strip() for v in line.split(":", 1)]
        metadata[key] = value
        line = f.readline()

    return url, content_type, encoding, metadata

# vim: tabstop=4 expandtab shiftwidth=4
